@@ -1,27 +1,21 @@
 use std::collections::HashMap;
-use std::fs;
+use std::fmt::Formatter;
 use std::io;
 use std::ops::Bound;
 use std::path::Path;
+use std::sync::Arc;
+use std::{fmt, fs};
 
 use anyhow::Result;
 
 use async_trait::async_trait;
 
-use chrono::{DateTime, Utc};
+use bytes::Bytes;
 
 use futures::Stream;
 
 use serde::{Deserialize, Serialize};
-
-use warp::hyper::body::Bytes;
-
-/// The type of a blob in a menmos cluster (file or directory).
-#[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
-pub enum Type {
-    File,
-    Directory,
-}
+use time::OffsetDateTime;
 
 fn file_to_base64<P: AsRef<Path>>(path: P) -> Result<String> {
     Ok(base64::encode(fs::read(path.as_ref())?))
@@ -47,53 +41,24 @@ impl CertificateInfo {
 }
 
 /// Metadata accepted when indexing a blob.
-#[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
+#[derive(Clone, Debug, Default, Deserialize, Serialize, PartialEq)]
 #[serde(deny_unknown_fields)]
 pub struct BlobMetaRequest {
-    /// The name of this blob. Does not need to be unique.
-    pub name: String,
-
-    /// The type of this blob.
-    pub blob_type: Type,
-
     /// The key/value pairs for this blob.
-    pub metadata: HashMap<String, String>,
+    pub fields: HashMap<String, FieldValue>,
 
     /// The tags for this blob.
     pub tags: Vec<String>,
-
-    /// This blob's parent IDs.
-    pub parents: Vec<String>,
-
-    /// This blob's size, in bytes.
-    pub size: u64,
 }
 
 impl BlobMetaRequest {
-    pub fn new<S: Into<String>>(name: S, blob_type: Type) -> Self {
-        Self {
-            name: name.into(),
-            blob_type,
-            metadata: Default::default(),
-            tags: Default::default(),
-            parents: Default::default(),
-            size: 0,
-        }
+    pub fn new() -> Self {
+        Self::default()
     }
 
     #[must_use]
-    pub fn file<S: Into<String>>(name: S) -> Self {
-        Self::new(name, Type::File)
-    }
-
-    #[must_use]
-    pub fn directory<S: Into<String>>(name: S) -> Self {
-        Self::new(name, Type::Directory)
-    }
-
-    #[must_use]
-    pub fn with_meta<S: Into<String>, T: Into<String>>(mut self, key: S, value: T) -> Self {
-        self.metadata.insert(key.into(), value.into());
+    pub fn with_field<S: Into<String>, T: Into<FieldValue>>(mut self, key: S, value: T) -> Self {
+        self.fields.insert(key.into(), value.into());
         self
     }
 
@@ -103,101 +68,137 @@ impl BlobMetaRequest {
         self
     }
 
-    #[must_use]
-    pub fn with_parent<S: Into<String>>(mut self, s: S) -> Self {
-        self.parents.push(s.into());
-        self
-    }
-
-    #[must_use]
-    pub fn with_size(mut self, size: u64) -> Self {
-        self.size = size;
-        self
-    }
-
-    pub fn into_meta(self, created_at: DateTime<Utc>, modified_at: DateTime<Utc>) -> BlobMeta {
+    pub fn into_meta(
+        self,
+        created_at: OffsetDateTime,
+        modified_at: OffsetDateTime,
+        size: u64,
+    ) -> BlobMeta {
         BlobMeta {
-            name: self.name,
-            blob_type: self.blob_type,
-            metadata: self.metadata,
+            fields: self.fields,
             tags: self.tags,
-            parents: self.parents,
-            size: self.size,
+            size,
             created_at,
             modified_at,
         }
     }
 }
 
+#[derive(Clone, Debug, Deserialize, Hash, Serialize, PartialEq, Eq)]
+#[serde(untagged)]
+pub enum FieldValue {
+    Str(String),
+    Numeric(i64),
+}
+
+#[derive(Clone, Debug, Deserialize, Hash, Serialize, PartialEq, Eq)]
+pub enum TaggedFieldValue {
+    Str(String),
+    Numeric(i64),
+}
+
+impl From<FieldValue> for TaggedFieldValue {
+    fn from(v: FieldValue) -> Self {
+        match v {
+            FieldValue::Str(s) => TaggedFieldValue::Str(s),
+            FieldValue::Numeric(i) => TaggedFieldValue::Numeric(i),
+        }
+    }
+}
+
+impl From<TaggedFieldValue> for FieldValue {
+    fn from(v: TaggedFieldValue) -> Self {
+        match v {
+            TaggedFieldValue::Str(s) => FieldValue::Str(s),
+            TaggedFieldValue::Numeric(i) => FieldValue::Numeric(i),
+        }
+    }
+}
+
+impl From<String> for FieldValue {
+    fn from(v: String) -> Self {
+        Self::Str(v)
+    }
+}
+
+impl From<&str> for FieldValue {
+    fn from(v: &str) -> Self {
+        Self::Str(String::from(v))
+    }
+}
+
+impl From<&String> for FieldValue {
+    fn from(v: &String) -> Self {
+        Self::Str(String::from(v))
+    }
+}
+
+impl From<i64> for FieldValue {
+    fn from(v: i64) -> Self {
+        Self::Numeric(v)
+    }
+}
+
+impl fmt::Display for FieldValue {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::Str(s) => write!(f, "\"{}\"", s),
+            Self::Numeric(i) => write!(f, "{}", i),
+        }
+    }
+}
+
 /// Metadata associated with a blob.
 #[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
 #[serde(deny_unknown_fields)]
 pub struct BlobMeta {
-    /// The name of this blob. Does not need to be unique.
-    pub name: String,
-
-    /// The type of this blob.
-    pub blob_type: Type,
-
     /// The key/value pairs for this blob.
-    pub metadata: HashMap<String, String>,
+    pub fields: HashMap<String, FieldValue>,
 
     /// The tags for this blob.
     pub tags: Vec<String>,
 
-    /// This blob's parent IDs.
-    pub parents: Vec<String>,
-
     /// This blob's size, in bytes.
     pub size: u64,
 
     /// This blob's creation time.
-    pub created_at: DateTime<Utc>,
+    #[serde(with = "time::serde::rfc3339")]
+    pub created_at: OffsetDateTime,
 
     /// This blob's last modified time.
-    pub modified_at: DateTime<Utc>,
+    #[serde(with = "time::serde::rfc3339")]
+    pub modified_at: OffsetDateTime,
 }
 
 impl From<BlobMeta> for BlobMetaRequest {
     fn from(m: BlobMeta) -> Self {
         Self {
-            name: m.name,
-            blob_type: m.blob_type,
-            metadata: m.metadata,
+            fields: m.fields,
             tags: m.tags,
-            parents: m.parents,
-            size: m.size,
+        }
+    }
+}
+
+impl Default for BlobMeta {
+    fn default() -> Self {
+        Self {
+            fields: Default::default(),
+            tags: Default::default(),
+            size: 0,
+            created_at: OffsetDateTime::now_utc(),
+            modified_at: OffsetDateTime::now_utc(),
         }
     }
 }
 
 impl BlobMeta {
-    pub fn new<S: Into<String>>(name: S, blob_type: Type) -> Self {
-        Self {
-            name: name.into(),
-            blob_type,
-            metadata: Default::default(),
-            tags: Default::default(),
-            parents: Default::default(),
-            size: 0,
-            created_at: Utc::now(),
-            modified_at: Utc::now(),
-        }
+    pub fn new() -> Self {
+        Self::default()
     }
 
     #[must_use]
-    pub fn file<S: Into<String>>(name: S) -> Self {
-        BlobMeta::new(name, Type::File)
-    }
-
-    #[must_use]
-    pub fn directory<S: Into<String>>(name: S) -> Self {
-        BlobMeta::new(name, Type::Directory)
-    }
-
-    #[must_use]
-    pub fn with_meta<S: Into<String>, T: Into<String>>(mut self, key: S, value: T) -> Self {
-        self.metadata.insert(key.into(), value.into());
+    pub fn with_field<S: Into<String>, T: Into<FieldValue>>(mut self, key: S, value: T) -> Self {
+        self.fields.insert(key.into(), value.into());
         self
     }
 
@@ -207,12 +208,6 @@ impl BlobMeta {
         self
     }
 
-    #[must_use]
-    pub fn with_parent<S: Into<String>>(mut self, s: S) -> Self {
-        self.parents.push(s.into());
-        self
-    }
-
     #[must_use]
     pub fn with_size(mut self, size: u64) -> Self {
         self.size = size;
@@ -220,17 +215,81 @@ impl BlobMeta {
     }
 }
 
+/// Tagged version of the metadata associated with a blob.
+///
+/// This is used to persist the metadata in the sled tree.
+/// bincode doesn't like untagged enums, so we have to make a tagged alternative.
+#[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
+#[serde(deny_unknown_fields)]
+pub struct TaggedBlobMeta {
+    /// The key/value pairs for this blob.
+    pub fields: HashMap<String, TaggedFieldValue>,
+
+    /// The tags for this blob.
+    pub tags: Vec<String>,
+
+    /// This blob's size, in bytes.
+    pub size: u64,
+
+    /// This blob's creation time.
+    #[serde(with = "crate::timestamp_nanos")]
+    pub created_at: OffsetDateTime,
+
+    /// This blob's last modified time.
+    #[serde(with = "crate::timestamp_nanos")]
+    pub modified_at: OffsetDateTime,
+}
+
+impl From<BlobMeta> for TaggedBlobMeta {
+    fn from(m: BlobMeta) -> Self {
+        Self {
+            fields: m
+                .fields
+                .into_iter()
+                .map(|(k, v)| (k, v.into()))
+                .collect::<HashMap<_, _>>(),
+            tags: m.tags,
+            size: m.size,
+            created_at: m.created_at,
+            modified_at: m.modified_at,
+        }
+    }
+}
+
+impl From<TaggedBlobMeta> for BlobMeta {
+    fn from(m: TaggedBlobMeta) -> Self {
+        Self {
+            fields: m
+                .fields
+                .into_iter()
+                .map(|(k, v)| (k, v.into()))
+                .collect::<HashMap<_, _>>(),
+            tags: m.tags,
+            size: m.size,
+            created_at: m.created_at,
+            modified_at: m.modified_at,
+        }
+    }
+}
+
 #[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
 #[serde(deny_unknown_fields)]
 pub struct BlobInfoRequest {
     pub meta_request: BlobMetaRequest,
     pub owner: String,
+    pub size: u64,
 }
 
 impl BlobInfoRequest {
-    pub fn into_blob_info(self, created_at: DateTime<Utc>, modified_at: DateTime<Utc>) -> BlobInfo {
+    pub fn into_blob_info(
+        self,
+        created_at: OffsetDateTime,
+        modified_at: OffsetDateTime,
+    ) -> BlobInfo {
         BlobInfo {
-            meta: self.meta_request.into_meta(created_at, modified_at),
+            meta: self
+                .meta_request
+                .into_meta(created_at, modified_at, self.size),
             owner: self.owner,
         }
     }
@@ -243,6 +302,31 @@ pub struct BlobInfo {
     pub owner: String,
 }
 
+#[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
+#[serde(deny_unknown_fields)]
+pub struct TaggedBlobInfo {
+    pub meta: TaggedBlobMeta,
+    pub owner: String,
+}
+
+impl From<BlobInfo> for TaggedBlobInfo {
+    fn from(v: BlobInfo) -> Self {
+        Self {
+            meta: v.meta.into(),
+            owner: v.owner,
+        }
+    }
+}
+
+impl From<TaggedBlobInfo> for BlobInfo {
+    fn from(v: TaggedBlobInfo) -> Self {
+        Self {
+            meta: v.meta.into(),
+            owner: v.owner,
+        }
+    }
+}
+
 pub struct Blob {
     pub stream: Box<dyn Stream<Item = Result<Bytes, io::Error>> + Send>,
     pub current_chunk_size: u64,
@@ -279,3 +363,5 @@ pub trait StorageNode {
 
     async fn flush(&self) -> Result<()>;
 }
+
+pub type DynStorageNode = Arc<dyn StorageNode + Send + Sync>;
